/* 
 * This program simulates a game of tennis. Each player is a process.
 *
 * Player 0 always serves. The ball is in play until one of the
 * players misses or hits it out. When one player reaces 5 points the
 * game is over.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include "ipc.h"

/* these aren't declared properly in the headers */
extern void srand48(int);
extern double drand48(void);

/*
 * declarations of functions in this file
 */
int hit(void);
int game_over(semid_t s, int *score);
int player(int n, semid_t ball_sem, semid_t score_sem, int *score, int *ball);
void clean_ipc(void);
void clean_exit();
void reaper();

/* Phrases describing the different 'hits' */
char *positions[]={
  "misses the ball!",
  "hits the ball out!", 
  "hits the net!", 
  "volleys", 
  "lobs",
  "hits to backcourt", 
  "hits crosscourt", 
  "hits to backcourt", 
  "hits crosscourt", 
  "hits to right",
  "hits to center",
  "hits to left",
  "hits to right",
  "hits to center",
  "hits to left"
};

/* how many of the above hits are "out" */
const int numouts=3; 

/* how many different hits there are */
const int numpos=(sizeof(positions) / sizeof(*positions));

int *score;
int *ball;

semid_t ball_sem;
semid_t score_sem;

int main(int argc, char **argv)
{
  int i;
  int n=0;

  /* this prevents shared memory from being left behind in case
     program is interrupted */
  signal(SIGINT,clean_exit);

  /* both players need to see the score */
  /* array of 2 numbers */
  if ((score=(int *)shmalloc(2*sizeof(*score))) == NULL) {
    fprintf(stderr,"can't allocate shared memory!\n"); /* failed */
    clean_exit();
  }

  if ((ball=(int *)shmalloc(sizeof(*ball))) == NULL) {
    fprintf(stderr,"can't allocate shared memory!\n"); /* failed */
    clean_exit();
  }

  /*
   * noone holds the ball initially. Actually, we start with no ball
   * at all, but we will 'create' one shortly...
   */
  if ((ball_sem=semcreate(0)) == -1) {
    perror("semcreate"); /* failed */
    clean_exit();
  }
  
  /* one player at a time is allowed to modify the score */
  if ((score_sem=semcreate(1)) == -1) {
    perror("semcreate"); /* failed */
    clean_exit();
  }


  /* zero the score */
  /* protection not necessary yet, we're still only one process */
  score[0]=score[1]=0;

  /* start two players */
  for (i=0; i<2; i++) {
    switch (fork()) {
    case -1:
      /* fork failed */
      perror("fork");
      
      /* this forces existing process to quit */
      /* (i.e. game is over) */
      score[0]=5;

      /* we clean up ipc and exit outside the switch */
      break;

    case 0:
      /* fork succeeded: child process */
      player(i, ball_sem, score_sem, score, ball);

      /* children exit here */
      return 0;

    default:
      /* fork succeeded: parent process */
      /* count the child */
      n++;
    }
  }

  /* if two successful forks */
  if (n==2) {
    /* start the game, making sure children had time to start */
    sleep(1);
    fprintf(stderr,"\nGame is starting...\n");
    fprintf(stderr,"score: 0 - 0\n\n");
    sleep(1);

    /* children are waiting, now 'create' the ball (starts game) */
    v(ball_sem); 
  }

  /* wait for children to finish and clean up after them */
  for (i=0; i<n; i++)
    wait(NULL);

  clean_ipc();

  return 0;
}


/*
 * This function represents both players.
 */
int player(int n, semid_t ball_sem, semid_t score_sem, int *score, int *ball)
{
  int s=(n+1)*20; /* fudge to get output in columns */

  /* 
   * set up the random number generator (if we do this before the
   * fork, then each player does exactly the same thing, so we do it
   * here to make them unique).
   */
  srand48(getpid());

  fprintf(stderr,"%*s%d ready\n",s,"",n);

  while (!game_over(score_sem,score)) {
    /* player 0 always serves */
    if (n==0) {
      /* get ball */
      p(ball_sem); 

      fprintf(stderr,"%*s%d serves\n",s,"",n);

      /* force the ball to be 'in' */
      *ball=numouts+1;

      sleep(1);

      /* hand ball to other player */
      v(ball_sem); 
    }

    /* ball is now in play */
    while (1) {
      /* get ball */
      p(ball_sem);  
      if (*ball < numouts) {
	/* last ball was out - exit 'in-play' loop */
	/* release ball */
	v(ball_sem); 
	break;
      }
      else {
	/* hit the ball */
	*ball=hit();
	fprintf(stderr,"%*s%d %s\n",s,"",n,positions[*ball]);
	if (*ball < numouts) {
	  /* 
	   * This player hit the ball out, he updates the score.
	   * Other player gets point and we go to next serve.
	   */
	  score[1-n] += 1;
	  fprintf(stderr,"score: %d - %d\n\n",score[0],score[1]);
	  sleep(1);
	}
	/* release ball */
	v(ball_sem);  
      }
    }
  }
  return 0;
}


/* 
 * Game is over when one player has 5 (yeah yeah I know how to score
 * tennis but this is just an example)
 */
int game_over(semid_t s, int *score)
{
  int r=0;

  p(s);  /* protect score */
  if (score[0] > 4) r=1;
  else if (score[1] > 4) r=1;
  v(s); 

  return r;
}

/* 
 * hit the ball to one of the positions described above 
 */
int hit(void)
{
  /* let this take a little time */
  sleep(1);
  /* where to hit the ball (see 'positions' above) */
  return (int)(drand48() * numpos);
}

/* clean up any ipc we have allocated */
void clean_ipc(void)
{
  if (ball) shfree(ball);
  if (score) shfree(score);
  if (ball_sem != -1) semdestroy(ball_sem);
  if (score_sem != -1) semdestroy(score_sem);
}

void clean_exit()
{
  clean_ipc();
  exit(1);
}
